// file      : xsd/cxx/parser/elements.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#ifndef CXX_PARSER_ELEMENTS_HXX
#define CXX_PARSER_ELEMENTS_HXX

#include <sstream>

#include <cxx/elements.hxx>

#include <cxx/parser/cli.hxx>

namespace CXX
{
  namespace Parser
  {
    //
    //
    class Context: public CXX::Context
    {
    public:
      Context (std::wostream& o, CLI::Options const& ops)
          : CXX::Context (o,
                          ops.value<CLI::char_type> (),
                          ops.value<CLI::include_with_brackets> (),
                          ops.value<CLI::include_prefix> (),
                          ops.value<CLI::export_symbol> (),
                          ops.value<CLI::namespace_map> (),
                          ops.value<CLI::namespace_regex> (),
                          ops.value<CLI::namespace_regex_trace> (),
                          ops.value<CLI::generate_inline> ()),
            xml_parser (xml_parser_),
            parser_template (parser_template_),
            xml_parser_ (ops.value<CLI::xml_parser> ()),
            parser_template_ (L"::xsd::cxx::parser::"
                              + xml_parser
                              + L"::parser")
      {
      }

    protected:
      Context (Context& c)
          : CXX::Context (c),
            xml_parser (c.xml_parser),
            parser_template (c.parser_template)
      {
      }

      Context (Context& c, std::wostream& o)
          : CXX::Context (c, o),
            xml_parser (c.xml_parser),
            parser_template (c.parser_template)
      {
      }

    public:
      String& xml_parser;
      String& parser_template;

    private:
      String xml_parser_;
      String parser_template_;
    };


    struct HasTemplateParam: Traversal::Complex,
                             Traversal::List,
                             Traversal::Member,
                             protected virtual Context
    {
      HasTemplateParam (Context& c, Boolean& v)
          : Context (c), v_ (v)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> *this;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        if (!v_)
          Traversal::Complex::traverse (c);
      }

      virtual Void
      traverse (SemanticGraph::List&)
      {
        v_ = true;
      }

      virtual Void
      traverse (SemanticGraph::Member& m)
      {
        if (!skip (m))
          v_ = true;
      }

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;

      Boolean& v_;
    };


    //
    //
    struct TemplateParamMemberDecl : Traversal::Member,
                                     protected virtual Context
    {
      TemplateParamMemberDecl (Context& c, Boolean& first)
          : Context (c), first_ (first)
      {
      }

      virtual Void
      traverse (Type& m)
      {
        if (skip (m))
          return;

        if (first_)
          first_ = false;
        else
          os << ", ";

        os << "typename _xsd_" << ename (m) << "_";
      }

    private:
      Boolean& first_;
    };


    //
    //
    struct TemplateParamBaseDecl : Traversal::Complex,
                                   Traversal::List,
                                   protected virtual Context
    {
      TemplateParamBaseDecl (Context& c, Boolean& first)
          : Context (c), first_ (first), member_ (c, first_)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> member_;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        //os << ", typename _xsd_" << ename (c) << "_";

        Traversal::Complex::traverse (c);
      }

      virtual Void
      traverse (SemanticGraph::List& l)
      {
        if (first_)
          first_ = false;
        else
          os << ", ";

        os << "typename _xsd_" << ename (l) << "_item_";
      }

    private:
      Boolean& first_;

      Traversal::Inherits inherits_;

      TemplateParamMemberDecl member_;
      Traversal::Names names_;
    };


    //
    //
    struct TemplateParamMemberUse : Traversal::Member,
                                    protected virtual Context
    {
      TemplateParamMemberUse (Context& c, Boolean& first)
          : Context (c), first_ (first)
      {
      }

      virtual Void
      traverse (Type& m)
      {
        if (skip (m))
          return;

        if (first_)
        {
          first_ = false;
          os << "< ";
        }
        else
          os << ", ";

        os << "_xsd_" << ename (m) << "_";
      }

    private:
      Boolean& first_;
    };


    //
    //
    struct TemplateParamBaseUse : Traversal::Complex,
                                  Traversal::List,
                                  protected virtual Context
    {
      TemplateParamBaseUse (Context& c, Boolean& first)
          : Context (c), first_ (first), member_ (c, first)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> member_;
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        //os << ", _xsd_" << ename (c) << "_";

        Traversal::Complex::traverse (c);
      }

      virtual Void
      traverse (SemanticGraph::List& l)
      {
        if (first_)
        {
          first_ = false;
          os << "< ";
        }
        else
          os << ", ";

        os << "_xsd_" << ename (l) << "_item_";
      }

    private:
      Boolean& first_;

      Traversal::Inherits inherits_;

      TemplateParamMemberUse member_;
      Traversal::Names names_;
    };


    //
    //
    struct TemplateParamDeclList: Traversal::Complex,
                                  protected virtual Context
    {
      TemplateParamDeclList (Context& c, Boolean first = true)
          : Context (c),
            first_ (first),
            template_param_member_ (c, first_),
            template_param_base_ (c, first_)
      {
        template_param_names_ >> template_param_member_;
        template_param_inherits_ >> template_param_base_;
      }

      virtual Void
      traverse (Type& c)
      {
        //os << "typename _xsd_" << ename (c) << "_";

        inherits (c, template_param_inherits_);
        names (c, template_param_names_);
      }

    private:
      Boolean first_;

      TemplateParamMemberDecl template_param_member_;
      Traversal::Names template_param_names_;

      TemplateParamBaseDecl template_param_base_;
      Traversal::Inherits template_param_inherits_;
    };


    //
    //
    struct TemplateParamUseList: Traversal::Complex,
                                 protected virtual Context
    {
      TemplateParamUseList (Context& c)
          : Context (c),
            first_ (true),
            template_param_member_ (c, first_),
            template_param_base_ (c, first_)
      {
        template_param_names_ >> template_param_member_;
        template_param_inherits_ >> template_param_base_;
      }

      ~TemplateParamUseList ()
      {
        if (!first_)
          os << " >";
      }

      virtual Void
      traverse (Type& c)
      {
        // os << "_xsd_" << ename (c) << "_";

        inherits (c, template_param_inherits_);
        names (c, template_param_names_);
      }

    private:
      Boolean first_;

      TemplateParamMemberUse template_param_member_;
      Traversal::Names template_param_names_;

      TemplateParamBaseUse template_param_base_;
      Traversal::Inherits template_param_inherits_;
    };


    //
    //
    struct ParserParamDecl : Traversal::Complex,
                             Traversal::List,
                             Traversal::Member,
                             protected virtual Context
    {
      ParserParamDecl (Context& c, Boolean name_arg)
          : Context (c), first_ (true), name_arg_ (name_arg)
      {
        *this >> inherits_ >> *this;
        *this >> names_ >> *this;
      }

      using Traversal::Complex::traverse;

      virtual Void
      traverse (SemanticGraph::List& l)
      {
        String const& name (ename (l));

        if (!first_)
          os << "," << endl;
        else
          first_ = false;

        os << parser_template << "< _xsd_" << name << "_item_, " <<
          char_type << " >&";

        if (name_arg_)
          os << " " << name << "_item";
      }

      virtual Void
      traverse (SemanticGraph::Member& m)
      {
        if (skip (m)) return;

        String const& name (ename (m));

        if (!first_)
          os << "," << endl;
        else
          first_ = false;

        os << parser_template << "< _xsd_" << name << "_, " <<
          char_type << " >&";

        if (name_arg_)
          os << " " << name;
      }

    private:
      Traversal::Inherits inherits_;
      Traversal::Names names_;

      Boolean first_;
      Boolean name_arg_;
    };


    //
    //
    struct Base : Traversal::Complex,
                  Traversal::Union,
                  Traversal::List,
                  protected virtual Context
    {
      Base (Context& c, String const& name, Boolean v)
          : Context (c),
            name_ (name),
            virtual_ (v),
            first_ (true),
            base_ (c, first_),
            member_ (c, first_)
      {
        *this >> inherits_ >> base_;
        *this >> names_ >> member_;
      }

      ~Base ()
      {
        if (!first_)
          os << " >";
      }

      virtual Void
      traverse (SemanticGraph::Complex& c)
      {
        String const& name (ename (c));
        String scope (fq_name (c.scope ()));

        os << (virtual_ ? "virtual " : "") <<
          scope << "::_xsd_" << name << "::" << name;

        Traversal::Complex::traverse (c);
      }

      virtual Void
      traverse (SemanticGraph::Union& u)
      {
        //String type (fq_name (u));
        //os << (virtual_ ? "virtual " : "") << type;

        os << (virtual_ ? "virtual " : "") <<
          "::xsd::cxx::parser::event_consumer_impl< " << char_type << " >";
      }

      virtual Void
      traverse (SemanticGraph::List& l)
      {
        String const& name (ename (l));
        String scope (fq_name (l.scope ()));

        //@@ The _item_ template parameter may conflict. There is
        //   not much I can do here. It seems the only way to handle
        //   this properly is to assign the name in the name processor.
        //
        os << (virtual_ ? "virtual " : "") <<
          scope << "::_xsd_" << name << "::" << name <<
          "< _xsd_" << name << "_item_ >";
      }

    protected:
      String name_;
      Boolean virtual_;
      Boolean first_;

    private:
      TemplateParamBaseUse base_;
      Traversal::Inherits inherits_;

      TemplateParamMemberUse member_;
      Traversal::Names names_;
    };

    //
    //
    struct BaseType: Base,
                     Traversal::Type
    {
      BaseType (Context& c, String const& name, Boolean v)
          : Context (c), Base (c, name, v)
      {
      }

      virtual Void
      traverse (SemanticGraph::Type& t)
      {
        os << (virtual_ ? "virtual " : "") <<
          "::xsd::cxx::parser::event_consumer_impl< " << char_type << " >";
      }
    };
  }
}

#endif  // CXX_PARSER_ELEMENTS_HXX
