// file      : xsd/xsd.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cult/types.hxx>

#include <cult/trace/log.hxx>

#include <cult/cli/exceptions.hxx>
#include <cult/cli/arguments.hxx>
#include <cult/cli/options.hxx>
#include <cult/cli/options-spec.hxx>
#include <cult/cli/options-parser.hxx>

#include <xsd-frontend/parser.hxx>

#include <cxx/parser/generator.hxx>
#include <cxx/tree/generator.hxx>

//#include <idl/generator.hxx>
//#include <idl/cxx/generator.hxx>

//#include <cxx-cli/parser/validator.hxx>
//#include <cxx-cli/parser/generator.hxx>

#include <morphing/anonymous/processor.hxx>

#include <processing/cardinality/processor.hxx>
#include <processing/inheritance/processor.hxx>

#include <iostream>

#include "usage.hxx"
#include "version.hxx"

using namespace Cult::Types;
using namespace XSDFrontend;

using std::wcerr;
using std::endl;

namespace CLI
{
  using namespace Cult::CLI;

  extern Char const help[]    = "help";
  extern Char const version[] = "version";

  typedef Cult::CLI::Options
  <
    help, Boolean,
    version, Boolean
  >
  Options;

  typedef
  Cult::CLI::OptionsSpec<Options>
  OptionsSpec;
}

Int
main (Int argc, Char* argv[])
{
  std::wostream& e (wcerr);

  Cult::Trace::Log::instance ().level (0);

  try
  {
    CLI::OptionsSpec spec;

    CLI::Options options (
      CLI::parse (spec, argc, argv, CLI::UnknownMode::stop));

    CLI::Arguments args (argc, argv);

    String cmd;

    if (args.size () > 1)
    {
      cmd = args[1];
      args.erase (1);
    }

    if (options.value<CLI::version> () || cmd == "version")
    {
      e << "XML Schema Definition Compiler "
        << version_major  << "."
        << version_minor  << "."
        << version_bugfix << endl
        << "Copyright (C) 2005-2006 Code Synthesis Tools CC" << endl
        << "This is free software; see the source for copying conditions. "
        << "There is NO\nwarranty; not even for MERCHANTABILITY or FITNESS "
        << "FOR A PARTICULAR PURPOSE." << endl;

      return 0;
    }

    if (options.value<CLI::help> () || cmd == "help")
    {
      if (cmd == "help" && args.size () > 1)
      {
        String arg (args[1]);

        if (arg == "cxx-tree")
        {
          e << "Usage: " << argv[0] << " cxx-tree [options] file" << endl
            << "Options:" << endl;

          CXX::Tree::Generator::usage ();
          Morphing::Anonymous::Processor::usage ();
        }
        else if (arg == "cxx-parser" || arg == "cxx-parser-e")
        {
          e << "Usage: " << argv[0] << " " << arg.c_str () <<
            " [options] file" << endl
            << "Options:" << endl;

          CXX::Parser::Generator::usage (arg);
          Morphing::Anonymous::Processor::usage ();
        }
        /*
        else if (arg == "idl")
        {
          IDL::Generator::usage ();
        }
        else if (arg == "idl-cxx")
        {
          IDL::CXX::Generator::usage ();
        }

        else if (arg == "cxx-cli-parser")
        {
          e << "Usage: " << argv[0] << " cxx--cli-parser [options] file" << endl
            << "Options:" << endl;

          CXX::Parser::Generator::usage (arg);
          Morphing::Anonymous::Processor::usage ();
        }
        */
        else
        {
          e << "unknown command '" << arg.c_str () << "'" << endl
            << "try '" << argv[0] << " help' for the list of commands"
            << endl;

          return 1;
        }
      }
      else
      {
        e << "Usage: " << argv[0] << " <cmd> ..." << endl
          << "Commands:" << endl;

        e << "  help            Print usage information and exit. Use\n"
          << "                  'help <cmd>' for command-specific options."
          << endl;


        e << "  version         Print version and exit."
          << endl;


        e << "  cxx-tree        Generate C++ tree mapping."
          << endl;


        e << "  cxx-parser      Generate C++ parser mapping."
          << endl;

        e << "  cxx-parser-e    Generate embedded C++ parser mapping."
          << endl;

        /*
        e << "  idl        CORBA IDL mapping."
          << endl;

        e << "  idl-cxx    C++ co-mapping for CORBA IDL mapping."
          << endl;

        e << "  cxx-cli-parser  Generate C++/CLI parser mapping."
          << endl;
        */
      }

      return 0;
    }

    if (cmd.empty ())
    {
      e << "no command specified" << endl
        << "try '" << argv[0] << " help' for usage information" << endl;

      return 1;
    }

    if (cmd != "cxx-tree" &&
        cmd != "cxx-parser" &&
        cmd != "cxx-parser-e"
        /*
        cmd != "idl" &&
        cmd != "idl-cxx"
        cmd != "cxx-cli-parser"
        */)
    {
      e << "unknown command '" << cmd.c_str () << "'" << endl
        << "try '" << argv[0] << " help' for the list of commands"
        << endl;

      return 1;
    }

    // We need to parse command line options before we can get to
    // the arguments.
    //
    Morphing::Anonymous::CLI::Options morph_anon_options (
      CLI::parse (Morphing::Anonymous::Processor::options_spec (),
                  argc,
                  argv,
                  CLI::UnknownMode::skip,
                  CLI::UnknownMode::skip));

    Evptr<CXX::Tree::CLI::Options> cxx_tree_options;
    Evptr<CXX::Parser::CLI::Options> cxx_parser_options;
    /*
    Evptr<CXX_CLI::Parser::CLI::Options> cxx_cli_parser_options;
    */

    if (cmd == "cxx-tree")
    {
      cxx_tree_options = new CXX::Tree::CLI::Options (
        CLI::parse (CXX::Tree::Generator::options_spec (), argc, argv));
    }
    else if (cmd == "cxx-parser" || cmd == "cxx-parser-e")
    {
      cxx_parser_options = new CXX::Parser::CLI::Options (
        CLI::parse (CXX::Parser::Generator::options_spec (cmd),
                    argc,
                    argv));
    }
    /*
    else if (cmd == "cxx-cli-parser")
    {
      cxx_cli_parser_options = new CXX_CLI::Parser::CLI::Options (
        CLI::parse (CXX_CLI::Parser::Generator::options_spec (),
                    argc,
                    argv));
    }
    */


    if (argc < 2)
    {
      e << argv[0] << ": error: no input file." << endl;

      return 1;
    }

    // Parse schema.
    //
    SemanticGraph::Path tu;

    try
    {
      tu = SemanticGraph::Path (argv[1], boost::filesystem::native);
    }
    catch (SemanticGraph::InvalidPath const&)
    {
      e << argv[0] << ": error: '" << argv[1] << "' is not a valid "
        << "filesystem path" << endl;

      return 1;
    }

    Parser parser;
    Evptr<SemanticGraph::Schema> schema (parser.parse (tu));


    // Morph anonymous types.
    //
    if (morph_anon_options.value<
          Morphing::Anonymous::CLI::morph_anonymous> ())
    {
      try
      {
        Morphing::Anonymous::Processor proc;
        proc.morph (morph_anon_options, *schema, tu);
      }
      catch (Morphing::Anonymous::Processor::Failed const&)
      {
        // Diagnostic has already been issued.
        //
        return 1;
      }
    }

    // Try to rearrange definitions so that there is no forward
    // inheritance.
    //
    {
      Processing::Inheritance::Processor proc;
      proc.process (*schema, tu);
    }

    // Calculate cardinality.
    //
    {
      Processing::Cardinality::Processor proc;
      proc.process (*schema, tu);
    }

    // Generate mapping.
    //
    if (cmd == "cxx-tree")
    {
      try
      {
        CXX::Tree::Generator generator;
        generator.generate (*cxx_tree_options, *schema, tu);
      }
      catch (CXX::Tree::Generator::Failed const&)
      {
        // Diagnostic has already been issued.
        //
        return 1;
      }
    }
    else if (cmd == "cxx-parser" || cmd == "cxx-parser-e")
    {
      try
      {
        CXX::Parser::Generator generator;
        generator.generate (*cxx_parser_options, *schema, tu);
      }
      catch (CXX::Parser::Generator::Failed const&)
      {
        // Diagnostic has already been issued.
        //
        return 1;
      }
    }
    else if (cmd == "idl")
    {
    }
    else if (cmd == "idl-cxx")
    {
    }
    /*
    else if (cmd == "cxx-cli-parser")
    {
      // Let the mapping evaluate the graph for possibility of
      // generating something useful.
      //
      {
        CXX_CLI::Parser::Validator validator;

        if (!validator.validate (*cxx_cli_parser_options, *schema, tu))
          return 1;
      }

      try
      {
        CXX_CLI::Parser::Generator generator;
        generator.generate (*cxx_cli_parser_options, *schema, tu);
      }
      catch (CXX_CLI::Parser::Generator::Failed const&)
      {
        // Diagnostic has already been issued.
        //
        return 1;
      }
    }
    */

    return 0;
  }
  catch (InvalidSchema const&)
  {
    // Diagnostic has already been issued.
  }
  catch (CLI::UnexpectedOption const& e)
  {
    wcerr << "unknown option '" << e.option ().c_str () << "'" << endl
          << "try '" << argv[0] << " help' for usage information" << endl;
  }
  catch (CLI::OptionFormat const& e)
  {
    wcerr << "value for option '" << e.option ().c_str () << "' is invalid"
          << endl
          << "try '" << argv[0] << " help' for usage information" << endl;
  }

  return 1;
}
